@@ -56,6 +56,7 @@
 #include "io_lib/cram.h"
 #include "io_lib/os.h"
 #include "io_lib/md5.h"
+#include "io_lib/crc32.h"
 
 //Whether CIGAR has just M or uses = and X to indicate match and mismatch
 //#define USE_X
@@ -146,8 +147,10 @@ cram_block_compression_hdr *cram_decode_compression_header(cram_fd *fd,
 	return NULL;
 
     if (b->method != RAW) {
-	if (cram_uncompress_block(b))
+	if (cram_uncompress_block(b)) {
+	    free(hdr);
 	    return NULL;
+	}
     }
 
     cp = (char *)b->data;
@@ -673,8 +676,12 @@ int cram_dependent_data_series(cram_fd *fd,
 	if (fd->required_fields & SAM_SEQ)
 	    hdr->data_series |= CRAM_SEQ;
 
+	if (!(fd->required_fields & SAM_AUX))
+	    // No easy way to get MD/NM without other tags at present
+	    fd->decode_md = 0;
+
 	if (fd->required_fields & SAM_QUAL)
-	    hdr->data_series |= CRAM_SEQ;
+	    hdr->data_series |= CRAM_QUAL;
 
 	if (fd->required_fields & SAM_AUX)
 	    hdr->data_series |= CRAM_RG | CRAM_TL | CRAM_aux;
@@ -928,6 +935,78 @@ int cram_dependent_data_series(cram_fd *fd,
     return 0;
 }
 
+/*
+ * Checks whether an external block is used solely by a single data series.
+ * Returns the codec type if so (EXTERNAL, BYTE_ARRAY_LEN, BYTE_ARRAY_STOP)
+ *         or 0 if not (E_NULL).
+ */
+static int cram_ds_unique(cram_block_compression_hdr *hdr, cram_codec *c,
+			  int id) {
+    int i, n_id = 0;
+    enum cram_encoding e_type = 0;
+
+    for (i = 0; i < DS_END; i++) {
+	cram_codec *c;
+	int bnum1, bnum2, old_n_id;
+
+	if (!(c = hdr->codecs[i]))
+	    continue;
+
+	bnum1 = cram_codec_to_id(c, &bnum2);
+
+	old_n_id = n_id;
+	if (bnum1 == id) {
+	    n_id++;
+	    e_type = c->codec;
+	}
+	if (bnum2 == id) {
+	    n_id++;
+	    e_type = c->codec;
+	}
+
+	if (n_id == old_n_id+2)
+	    n_id--; // len/val in same place counts once only.
+    }
+
+    return n_id == 1 ? e_type : 0;
+}
+
+/*
+ * Attempts to estimate the size of some blocks so we can preallocate them
+ * before decoding.  Although decoding will automatically grow the blocks,
+ * it is typically more efficient to preallocate.
+ */
+void cram_decode_estimate_sizes(cram_block_compression_hdr *hdr, cram_slice *s,
+				int *qual_size, int *name_size,
+				int *q_id) {
+    int bnum1, bnum2;
+    cram_codec *cd;
+
+    *qual_size = 0;
+    *name_size = 0;
+
+    /* Qual */
+    cd = hdr->codecs[DS_QS];
+    bnum1 = cram_codec_to_id(cd, &bnum2);
+    if (bnum1 < 0 && bnum2 >= 0) bnum1 = bnum2;
+    if (cram_ds_unique(hdr, cd, bnum1)) {
+	cram_block *b = cram_get_block_by_id(s, bnum1);
+	if (b) *qual_size = b->uncomp_size;
+	if (q_id && cd->codec == E_EXTERNAL)
+	    *q_id = bnum1;
+    }
+
+    /* Name */
+    cd = hdr->codecs[DS_RN];
+    bnum1 = cram_codec_to_id(cd, &bnum2);
+    if (bnum1 < 0 && bnum2 >= 0) bnum1 = bnum2;
+    if (cram_ds_unique(hdr, cd, bnum1)) {
+	cram_block *b = cram_get_block_by_id(s, bnum1);
+	if (b) *name_size = b->uncomp_size;
+    }
+}
+
+
 /* ----------------------------------------------------------------------
  * CRAM slices
  */
@@ -963,16 +1042,21 @@ cram_block_slice_hdr *cram_decode_slice_header(cram_fd *fd, cram_block *b) {
     hdr->content_type = b->content_type;
 
     if (b->content_type == MAPPED_SLICE) {
-        cp += safe_itf8_get((char *)cp, (char *) cp_end, &hdr->ref_seq_id);
-        cp += safe_itf8_get((char *)cp, (char *) cp_end, &hdr->ref_seq_start);
-        cp += safe_itf8_get((char *)cp, (char *) cp_end, &hdr->ref_seq_span);
+        cp += safe_itf8_get((char *)cp, (char *)cp_end, &hdr->ref_seq_id);
+        cp += safe_itf8_get((char *)cp, (char *)cp_end, &hdr->ref_seq_start);
+        cp += safe_itf8_get((char *)cp, (char *)cp_end, &hdr->ref_seq_span);
     }
     cp += safe_itf8_get((char *)cp, (char *) cp_end, &hdr->num_records);
-    if (!IS_CRAM_1_VERS(fd))
-	cp += safe_itf8_get((char *)cp, (char *) cp_end, &hdr->record_counter);
-    cp += safe_itf8_get((char *)cp, (char *) cp_end, &hdr->num_blocks);
-
-    cp += safe_itf8_get((char *)cp, (char *) cp_end, &hdr->num_content_ids);
+    hdr->record_counter = 0;
+    if (CRAM_MAJOR_VERS(fd->version) == 2) {
+	int32_t i32 = 0;
+	cp += safe_itf8_get((char *)cp, (char *)cp_end, &i32);
+	hdr->record_counter = i32;
+    } else if (CRAM_MAJOR_VERS(fd->version) >= 3) {
+	cp += safe_ltf8_get((char *)cp, (char *)cp_end, &hdr->record_counter);
+    }
+    cp += safe_itf8_get((char *)cp, (char *)cp_end, &hdr->num_blocks);
+    cp += safe_itf8_get((char *)cp, (char *)cp_end, &hdr->num_content_ids);
     if (hdr->num_content_ids < 1 ||
 	hdr->num_content_ids >= SIZE_MAX / sizeof(int32_t)) {
         /* Slice must have at least one data block,
@@ -1209,7 +1293,7 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
     int decode_nm = fd->decode_md && s->ref && !has_NM;
     uint32_t ds = c->comp_hdr->data_series;
 
-    if (ds & CRAM_QS && !(cf & CRAM_FLAG_PRESERVE_QUAL_SCORES)) {
+    if ((ds & CRAM_QS) && !(cf & CRAM_FLAG_PRESERVE_QUAL_SCORES)) {
 	memset(qual, 255, cr->len);
     }
 
@@ -1323,7 +1407,7 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 	prev_pos = pos;
 
 	if (!(ds & CRAM_FC))
-	    continue;
+	    goto skip_cigar;
 
 	switch(op) {
 	case 'S': { // soft clip: IN
@@ -1340,7 +1424,9 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		    r |= c->comp_hdr->codecs[DS_IN]
 			? c->comp_hdr->codecs[DS_IN]
 			             ->decode(s, c->comp_hdr->codecs[DS_IN],
-					      blk, &seq[pos-1], &out_sz2)
+					      blk,
+					      cr->len ? &seq[pos-1] : NULL,
+					      &out_sz2)
 			: (seq[pos-1] = 'N', out_sz2 = 1, 0);
 		    have_sc = 1;
 		}
@@ -1351,7 +1437,9 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		    r |= c->comp_hdr->codecs[DS_SC]
 			? c->comp_hdr->codecs[DS_SC]
 			             ->decode(s, c->comp_hdr->codecs[DS_SC],
-					      blk, &seq[pos-1], &out_sz2)
+					      blk,
+					      cr->len ? &seq[pos-1] : NULL,
+					      &out_sz2)
 			: (seq[pos-1] = 'N', out_sz2 = 1, 0);
 		    have_sc = 1;
 		}
@@ -1385,7 +1473,8 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		r |= c->comp_hdr->codecs[DS_BS]
 		                ->decode(s, c->comp_hdr->codecs[DS_BS], blk,
 					 (char *)&base, &out_sz);
-		seq[pos-1] = 'N'; // FIXME look up BS=base value
+		if (pos-1 < cr->len)
+		    seq[pos-1] = 'N'; // FIXME look up BS=base value
 	    }
 	    cig_op = BAM_CBASE_MISMATCH;
 #else
@@ -1401,8 +1490,9 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 					 (char *)&base, &out_sz);
 		if (r) return -1;
 		if (ref_pos >= bfd->ref[cr->ref_id].len || !s->ref) {
-                    seq[pos-1] = c->comp_hdr->
-                        substitution_matrix[fd->L1['N']][base];
+		    if (pos-1 < cr->len)
+			seq[pos-1] = c->comp_hdr->
+			    substitution_matrix[fd->L1['N']][base];
 		    if (decode_md || decode_nm) {
 			if (md_dist >= 0 && decode_md)
 			    BLOCK_APPEND_UINT(s->aux_blk, md_dist);
@@ -1414,8 +1504,9 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
                         ? (uc)s->ref[ref_pos - s->ref_start +1]
                         : 'N';
                     ref_base = fd->L1[ref_call];
-		    seq[pos-1] = c->comp_hdr->
-			substitution_matrix[ref_base][base];
+		    if (pos-1 < cr->len)
+			seq[pos-1] = c->comp_hdr->
+			    substitution_matrix[ref_base][base];
 		    if (decode_md) {
                         BLOCK_APPEND_UINT(s->aux_blk, md_dist);
                         BLOCK_APPEND_CHAR(s->aux_blk, ref_call);
@@ -1494,7 +1585,8 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		if (!c->comp_hdr->codecs[DS_IN]) return -1;
 		r |= c->comp_hdr->codecs[DS_IN]
 		                ->decode(s, c->comp_hdr->codecs[DS_IN], blk,
-					 &seq[pos-1], &out_sz2);
+					 cr->len ? &seq[pos-1] : NULL,
+					 &out_sz2);
 		if (r) return r;
 		cig_op = BAM_CINS;
 		cig_len += out_sz2;
@@ -1514,7 +1606,8 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		if (!c->comp_hdr->codecs[DS_BA]) return -1;
 		r |= c->comp_hdr->codecs[DS_BA]
 		                ->decode(s, c->comp_hdr->codecs[DS_BA], blk,
-					 (char *)&seq[pos-1], &out_sz);
+					 cr->len ? &seq[pos-1] : NULL,
+					 &out_sz);
 		if (r) return r;
 		//printf("  %d: BA = %c (ret %d)\n", f, seq[pos-1], r);
 	    }
@@ -1537,7 +1630,8 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		if (!c->comp_hdr->codecs[DS_BB]) return -1;
 		r |= c->comp_hdr->codecs[DS_BB]
 		    ->decode(s, c->comp_hdr->codecs[DS_BB], blk,
-			     (char *)&seq[pos-1], &len);
+			     cr->len ? &seq[pos-1] : NULL,
+			     &len);
 		if (r) return r;
 
 		if (decode_md || decode_nm) {
@@ -1613,7 +1707,8 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		if (!c->comp_hdr->codecs[DS_BA]) return -1;
 		r |= c->comp_hdr->codecs[DS_BA]
 		                ->decode(s, c->comp_hdr->codecs[DS_BA], blk,
-					 (char *)&seq[pos-1], &out_sz);
+					 cr->len ? &seq[pos-1] : NULL,
+					 &out_sz);
 
 		if (decode_md || decode_nm) {
 		    if (md_dist >= 0 && decode_md)
@@ -1722,7 +1817,7 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 	goto skip_cigar;
 
     /* An implicit match op for any unaccounted for bases */
-    if (ds & CRAM_FN && cr->len >= seq_pos) {
+    if ((ds & CRAM_FN) && cr->len >= seq_pos) {
 	if (s->ref) {
 	    if (ref_pos + cr->len - seq_pos + 1 > bfd->ref[cr->ref_id].len) {
 		static int whinged = 0;
@@ -1732,17 +1827,20 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		whinged = 1;
 		rlen = bfd->ref[cr->ref_id].len - ref_pos;
 		if (rlen > 0) {
-		    memcpy(&seq[seq_pos-1],
-			   &s->ref[ref_pos - s->ref_start +1], rlen);
+		    if (seq_pos-1 + rlen < cr->len)
+			memcpy(&seq[seq_pos-1],
+			       &s->ref[ref_pos - s->ref_start +1], rlen);
 		    if ((cr->len - seq_pos + 1) - rlen > 0)
 		        memset(&seq[seq_pos-1+rlen], 'N',
                                (cr->len - seq_pos + 1) - rlen);
 		} else {
-		    memset(&seq[seq_pos-1], 'N', cr->len - seq_pos + 1);
+		    if (cr->len - seq_pos + 1 > 0)
+			memset(&seq[seq_pos-1], 'N', cr->len - seq_pos + 1);
 		}
 	    } else {
-		memcpy(&seq[seq_pos-1], &s->ref[ref_pos - s->ref_start +1],
-		       cr->len - seq_pos + 1);
+		if (cr->len - seq_pos + 1 > 0)
+		    memcpy(&seq[seq_pos-1], &s->ref[ref_pos - s->ref_start +1],
+			   cr->len - seq_pos + 1);
 		ref_pos += cr->len - seq_pos + 1;
 		if (md_dist >= 0)
 		    md_dist += cr->len - seq_pos + 1;
@@ -2002,6 +2100,8 @@ static int cram_decode_slice_xref(cram_slice *s, int required_fields) {
 		    int aleft = cr->apos, aright = cr->aend;
 		    int tlen;
 		    int ref = cr->ref_id;
+
+		    // number of segments starting at the same point.
 		    int left_cnt = 0;
 
 		    do {
@@ -2171,6 +2271,27 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 
     blk->bit = 7; // MSB first
 
+    // Study the blocks and estimate approx sizes to preallocate.
+    // This looks to speed up decoding by around 8-9%.
+    // We can always shrink back down at the end if we overestimated.
+    // However it's likely that this also saves memory as own growth
+    // factor (*=1.5) is never applied.
+    {
+	int qsize, nsize, q_id;
+	cram_decode_estimate_sizes(c->comp_hdr, s, &qsize, &nsize, &q_id);
+	//fprintf(stderr, "qsize=%d nsize=%d\n", qsize, nsize);
+
+	if (qsize && (ds & CRAM_RL)) BLOCK_RESIZE_EXACT(s->seqs_blk, qsize+1);
+	if (qsize && (ds & CRAM_RL)) BLOCK_RESIZE_EXACT(s->qual_blk, qsize+1);
+	if (nsize && (ds & CRAM_NS)) BLOCK_RESIZE_EXACT(s->name_blk, nsize+1);
+
+	// To do - consider using q_id here to usurp the quality block and
+	// avoid a memcpy during decode.
+	// Specifically when quality is an external block uniquely used by
+	// DS_QS only, then we can set s->qual_blk directly to this
+	// block and save the codec->decode() calls. (Approx 3% cpu saving)
+    }
+
     /* Look for unknown RG, added as last by Java CRAM? */
     if (bfd->nrg > 0 &&
 	bfd->rg[bfd->nrg-1].name != NULL &&
@@ -2199,7 +2320,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    b = cram_get_block_by_id(s, s->hdr->ref_base_id);
 	    if (!b)
 		return -1;
-            if (cram_uncompress_block(b) < 0)
+            if (cram_uncompress_block(b) != 0)
                 return -1;
 	    s->ref = (char *)BLOCK_DATA(b);
 	    s->ref_start = s->hdr->ref_seq_start;
@@ -2302,7 +2423,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    return -1;
     }
 
-    int last_ref_id = -9;
+    int last_ref_id = -9; // Arbitrary -ve marker for not-yet-set
     for (rec = 0; rec < s->hdr->num_records; rec++) {
 	cram_record *cr = &s->crecs[rec];
 	int has_MD, has_NM;
@@ -2445,7 +2566,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	cr->mate_pos = 0;
 	cr->mate_line = -1;
 	cr->mate_ref_id = -1;
-	if (ds & CRAM_CF && cf & CRAM_FLAG_DETACHED) {
+	if ((ds & CRAM_CF) && (cf & CRAM_FLAG_DETACHED)) {
 	    if (ds & CRAM_MF) {
 		if (IS_CRAM_1_VERS(fd)) {
 		    /* MF is byte in 1.0, int32 in 2.0 */
@@ -2516,7 +2637,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    } else {
 		cr->tlen = INT_MIN;
 	    }
-	} else if (ds & CRAM_CF && cf & CRAM_FLAG_MATE_DOWNSTREAM) {
+	} else if ((ds & CRAM_CF) && (cf & CRAM_FLAG_MATE_DOWNSTREAM)) {
 	    if (ds & CRAM_NF) {
 		if (!c->comp_hdr->codecs[DS_NF]) return -1;
 		r |= c->comp_hdr->codecs[DS_NF]
@@ -2615,7 +2736,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 		if (r) return r;
 	    }
 
-	    if (ds & CRAM_CF && cf & CRAM_FLAG_PRESERVE_QUAL_SCORES) {
+	    if ((ds & CRAM_CF) && (cf & CRAM_FLAG_PRESERVE_QUAL_SCORES)) {
 		out_sz2 = cr->len;
 		if (ds & CRAM_QS && cr->len >= 0) {
 		    if (!c->comp_hdr->codecs[DS_QS]) return -1;
@@ -2632,12 +2753,12 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 
 	if (!fd->ignore_chksum) {
 	    if (s->hdr->BD_crc && ds & CRAM_BA && s->ref)
-		s->BD_crc += crc32(0L, (Bytef *) seq, cr->len);
+		s->BD_crc += iolib_crc32(0L, (Bytef *) seq, cr->len);
 	    
 	    if (s->hdr->SD_crc &&
 		(ds & CRAM_QS) &&
 		(cf & CRAM_FLAG_PRESERVE_QUAL_SCORES)) {
-		s->SD_crc += crc32(0L, (Bytef *) qual, cr->len);
+		s->SD_crc += iolib_crc32(0L, (Bytef *) qual, cr->len);
 	    }
 	}
     }
@@ -2658,6 +2779,30 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
     /* Resolve mate pair cross-references between recs within this slice */
     r |= cram_decode_slice_xref(s, fd->required_fields);
 
+    // Free the original blocks as we no longer need these.
+    {
+	int i;
+	for (i = 0; i < s->hdr->num_blocks; i++) {
+	    cram_block *b = s->block[i];
+	    cram_free_block(b);
+	    s->block[i] = NULL;
+	}
+    }
+
+    // Also see initial BLOCK_RESIZE_EXACT at top of function.
+    // As we grow blocks we overallocate by up to 50%. So shrink
+    // back to their final sizes here.
+    //
+//    fprintf(stderr, "%d %d // %d %d // %d %d // %d %d\n",
+//    	    (int)s->seqs_blk->byte, (int)s->seqs_blk->alloc, 
+//    	    (int)s->qual_blk->byte, (int)s->qual_blk->alloc, 
+//    	    (int)s->name_blk->byte, (int)s->name_blk->alloc, 
+//    	    (int)s->aux_blk->byte,  (int)s->aux_blk->alloc);
+    BLOCK_RESIZE_EXACT(s->seqs_blk, BLOCK_SIZE(s->seqs_blk)+1);
+    BLOCK_RESIZE_EXACT(s->qual_blk, BLOCK_SIZE(s->qual_blk)+1);
+    BLOCK_RESIZE_EXACT(s->name_blk, BLOCK_SIZE(s->name_blk)+1);
+    BLOCK_RESIZE_EXACT(s->aux_blk,  BLOCK_SIZE(s->aux_blk)+1);
+
     /* Checksum */
     if (CRAM_MAJOR_VERS(fd->version) >= 3) {
 	if (ds & CRAM_BA && s->BD_crc && s->hdr->BD_crc) {
@@ -2761,15 +2906,21 @@ static int cram_to_bam(SAM_hdr *bfd, cram_fd *fd, cram_slice *s,
 	    name = (char *)BLOCK_DATA(s->name_blk) + cr->name;
 	    name_len = cr->name_len;
 	} else {
-	    // FIXME: add prefix, container number, slice number, etc
 	    name = name_a;
-
+	    name_len = strlen(fd->prefix);
+	    memcpy(name, fd->prefix, name_len);
+	    name += name_len;
+	    *name++ = ':';
 	    if (cr->mate_line >= 0 && cr->mate_line < rec)
-		name_len = sprintf(name_a, "%s:%"PRId64":%d",
-				   fd->prefix, s->id, cr->mate_line);
+		name = (char *)append_uint64((unsigned char *)name,
+					     s->hdr->record_counter +
+					     cr->mate_line + 1);
 	    else
-		name_len = sprintf(name_a, "%s:%"PRId64":%d",
-				   fd->prefix, s->id, rec);
+		name = (char *)append_uint64((unsigned char *)name,
+					     s->hdr->record_counter +
+					     rec + 1);
+	    name_len = name - name_a;
+	    name = name_a;
 	}
     } else {
 	name = "?";
@@ -2934,18 +3085,18 @@ static cram_slice *cram_next_slice(cram_fd *fd, cram_container **cp) {
 		    fd->required_fields |= SAM_POS;
 
 		    if (c->ref_seq_id != fd->range.refid) {
-			fd->eof = 1;
 			cram_free_container(c);
 			fd->ctr = NULL;
 			fd->ooc = 1;
+			fd->eof = 1;
 			break;
 		    }
 
 		    if (c->ref_seq_start > fd->range.end) {
-			fd->eof = 1;
 			cram_free_container(c);
 			fd->ctr = NULL;
 			fd->ooc = 1;
+			fd->eof = 1;
 			break;
 		    }
 
@@ -3103,7 +3254,17 @@ cram_record *cram_get_seq(cram_fd *fd) {
 	}
 
 	if (fd->range.refid != -2) {
-	    if (s->crecs[c->curr_rec].ref_id < fd->range.refid) {
+	    if (fd->range.refid == -1 && s->crecs[c->curr_rec].ref_id != -1) {
+		// Special case when looking for unmapped blocks at end.
+		// If these are mixed in with mapped data (c->ref_id == -2)
+		// then we need skip until we find the unmapped data, if at all
+		c->curr_rec++;
+		continue;
+	    }
+	    if (s->crecs[c->curr_rec].ref_id < fd->range.refid &&
+		s->crecs[c->curr_rec].ref_id != -1) {
+		// Looking for a mapped read, but not there yet.  Special case
+		// as -1 (unmapped) shouldn't be considered < refid.
 		c->curr_rec++;
 		continue;
 	    }
@@ -3115,14 +3276,14 @@ cram_record *cram_get_seq(cram_fd *fd) {
 		return NULL;
 	    }
 
-	    if (s->crecs[c->curr_rec].apos > fd->range.end) {
+	    if (fd->range.refid != -1 && s->crecs[c->curr_rec].apos > fd->range.end) {
 		fd->eof = 1;
 		cram_free_slice(s);
 		c->slice = NULL;
 		return NULL;
 	    }
 
-	    if (s->crecs[c->curr_rec].aend < fd->range.start) {
+	    if (fd->range.refid != -1 && s->crecs[c->curr_rec].aend < fd->range.start) {
 		c->curr_rec++;
 		continue;
 	    }
