@@ -326,7 +326,7 @@ typedef struct {
     int32_t ref_seq_start;  /* if content_type == MAPPED_SLICE */
     int32_t ref_seq_span;   /* if content_type == MAPPED_SLICE */
     int32_t num_records;
-    int32_t record_counter;
+    int64_t record_counter;
     int32_t num_blocks;
     int32_t num_content_ids;
     int32_t *block_content_ids;
@@ -353,7 +353,7 @@ typedef struct {
     int32_t  ref_seq_id;
     int32_t  ref_seq_start;
     int32_t  ref_seq_span;
-    int32_t  record_counter;
+    int64_t  record_counter;
     int64_t  num_bases;
     int32_t  num_records;
     int32_t  num_blocks;
@@ -542,9 +542,6 @@ typedef struct cram_slice {
     /* State used during encoding/decoding */
     int last_apos, max_apos;
 
-    /* Identifier used for auto-assigning read names */
-    uint64_t id;
-
     /* Array of decoded cram records */
     cram_record *crecs;
 
@@ -749,8 +746,7 @@ typedef struct {
     SAM_hdr       *header;
 
     char          *prefix;
-    int            record_counter;
-    int            slice_num;
+    int64_t        record_counter;
     int            err;
 
     // Most recent compression header decoded
@@ -775,7 +771,6 @@ typedef struct {
     int level;
     cram_metrics *m[DS_END];
     HashTable *tags_used; // cram_metrics[], per tag types in use.
-    int next_content_id;
 
     // options
     int decode_md; // Whether to export MD and NM tags
@@ -822,6 +817,9 @@ typedef struct {
 
     int ooc;                            // out of containers.
     int ignore_chksum;
+    int lossy_read_names;
+    int preserve_aux_order;             // if set implies emitting RG, MD and NM
+    int preserve_aux_size;              // does not replace 'i' with 'c' etc in aux.
 } cram_fd;
 
 #if defined(CRAM_IO_CUSTOM_BUFFERING)
@@ -914,8 +912,10 @@ enum cram_fields {
 #define CRAM_CIGAR (CRAM_FN | CRAM_FP | CRAM_FC | CRAM_DL | CRAM_IN | \
 		    CRAM_SC | CRAM_HC | CRAM_PD | CRAM_RS | CRAM_RL | CRAM_BF)
 
-#define CRAM_SEQ (CRAM_CIGAR | CRAM_BA | CRAM_QS | CRAM_BS | \
-		  CRAM_RL    | CRAM_AP | CRAM_BB | CRAM_QQ)
+#define CRAM_SEQ (CRAM_CIGAR | CRAM_BA | CRAM_BS | \
+		  CRAM_RL    | CRAM_AP | CRAM_BB)
+
+#define CRAM_QUAL (CRAM_CIGAR | CRAM_RL | CRAM_AP | CRAM_QS | CRAM_QQ)
 
 enum cram_option {
     CRAM_OPT_DECODE_MD,
@@ -941,6 +941,9 @@ enum cram_option {
     CRAM_OPT_USE_RANS,
     CRAM_OPT_IGNORE_CHKSUM,
     CRAM_OPT_BASES_PER_SLICE,
+    CRAM_OPT_LOSSY_READ_NAMES,
+    CRAM_OPT_PRESERVE_AUX_ORDER,
+    CRAM_OPT_PRESERVE_AUX_SIZE
 };
 
 /* BF bitfields */
@@ -974,6 +977,11 @@ enum cram_option {
 #define CRAM_FLAG_DETACHED             (1<<1)
 #define CRAM_FLAG_MATE_DOWNSTREAM      (1<<2)
 #define CRAM_FLAG_NO_SEQ               (1<<3)
+#define CRAM_FLAG_MASK                 ((1<<4)-1)
+
+/* Internal only */
+#define CRAM_FLAG_STATS_ADDED  (1<<30)
+#define CRAM_FLAG_DISCARD_NAME (1<<31)
 
 #ifdef __cplusplus
 }
