@@ -45,6 +45,10 @@
 #include <unistd.h>
 #include <stdarg.h>
 
+#ifdef _MSC_VER
+#include <io.h>
+#endif
+
 #include "io_lib/os.h"
 #include "io_lib/mFILE.h"
 #include "io_lib/vlen.h"
@@ -53,6 +57,14 @@
 #include <sys/mman.h>
 #endif
 
+//Moved from mFILE.h to avoid conflicting with WinUser.h
+#define MF_READ    1
+#define MF_WRITE   2
+#define MF_APPEND  4
+#define MF_BINARY  8
+#define MF_TRUNC  16
+#define MF_MODEX  32
+#define MF_MMAP   64
 
 /*
  * This file contains memory-based versions of the most commonly used
@@ -339,8 +351,20 @@ mFILE *mfreopen(const char *path, const char *mode_str, FILE *fp) {
  */
 mFILE *mfopen(const char *path, const char *mode) {
     FILE *fp;
+    char mode2[11];
+    int i1 = 0, i2 = 0;
 
-    if (NULL == (fp = fopen(path, mode)))
+    /* Remove the 'm' mmap symbol from mode before calling fopen() as
+     * MS Visual Studio dislikes it remaining.
+     */
+    while (i1 < 10 && mode[i1]) {
+	if (mode[i1] != 'm')
+	    mode2[i2++] = mode[i1];
+	i1++;
+    }
+    mode2[i2] = 0;
+
+    if (NULL == (fp = fopen(path, mode2)))
 	return NULL;
     return mfreopen(path, mode, fp);
 }
