@@ -118,6 +118,10 @@ static void usage(FILE *fp) {
 #ifdef HAVE_LIBLZMA
     fprintf(fp, "    -Z             [Cram] Also compress using lzma.\n");
 #endif
+    fprintf(fp, "    -n             [Cram] Discard read names where possible.\n");
+    fprintf(fp, "    -P             [Cram EXPERIMENTAL] Preserve all aux tags (incl RG,NM,MD)\n");
+    fprintf(fp, "    -p             [Cram EXPERIMENTAL] Preserve aux tag sizes ('i', 's', 'c')\n");
+    fprintf(fp, "    -N integer     Stop decoding after 'integer' sequences\n");
     fprintf(fp, "    -t N           Use N threads (availability varies by format)\n");
     fprintf(fp, "    -B             Enable Illumina 8 quality-binning system (lossy)\n");
     fprintf(fp, "    -!             Disable all checking of checksums\n");
@@ -142,9 +146,12 @@ int main(int argc, char **argv) {
     int sam_fields = 0; // all
     int header = 1;
     int bases_per_slice = 0;
+    int lossy_read_names = 0;
+    int preserve_aux_order = 0;
+    int preserve_aux_size = 0;
 
     /* Parse command line arguments */
-    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xXeI:O:R:!MmjJZt:BN:F:Hb:")) != -1) {
+    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xXeI:O:R:!MmjJZt:BN:F:Hb:nPp")) != -1) {
 	switch (c) {
 	case 'F':
 	    sam_fields = strtol(optarg, NULL, 0); // undocumented for testing
@@ -241,6 +248,10 @@ int main(int argc, char **argv) {
 	    ignore_md5 = 1;
 	    break;
 
+	case 'n':
+	    lossy_read_names = 1;
+	    break;
+
 	case 'M':
 	    multi_seq = 1;
 	    break;
@@ -279,7 +290,15 @@ int main(int argc, char **argv) {
 	    binning = BINNING_ILLUMINA;
 	    break;
 
-	case 'N': // For debugging
+	case 'P':
+	    preserve_aux_order = 1;
+	    break;
+
+	case 'p':
+	    preserve_aux_size = 1;
+	    break;
+
+	case 'N':
 	    max_reads = atoi(optarg);
 	    break;
 
@@ -409,6 +428,19 @@ int main(int argc, char **argv) {
 	    return 1;
     }
     
+    if (lossy_read_names) {
+	if (scram_set_option(out, CRAM_OPT_LOSSY_READ_NAMES, lossy_read_names))
+	    return 1;
+    }
+
+    if (preserve_aux_order)
+	if (scram_set_option(out, CRAM_OPT_PRESERVE_AUX_ORDER, preserve_aux_order))
+	    return 1;
+
+    if (preserve_aux_size)
+	if (scram_set_option(out, CRAM_OPT_PRESERVE_AUX_SIZE, preserve_aux_size))
+	    return 1;
+
     if (sam_fields)
 	scram_set_option(in, CRAM_OPT_REQUIRED_FIELDS, sam_fields);
 
